"""The client module handles the Redis Client interface.

Classes:
    RedisClient
"""
from collections import defaultdict
import typing as t

from respy3 import protocol

from . import connections


PUSH_COMMANDS = {b"SUBSCRIBE", b"PSUBSCRIBE", b"UNSUBSCRIBE", b"PUNSUBSCRIBE"}


class RedisClient:
    """RedisClient communicates with the Redis server via its *call* method.

    Attributes:
        host (str): The address to connect to (default: "127.0.0.1").
        port (int): The port to connect to (default: 6379).
        connection_pool (instance of a connection pool): The pool to use for
            connections. Leave as None to use the default ConnectionPool.
        Reader (protocol class): The class to use for interpreting responses from Redis.
        write_command (function): The function to use to format commands to send
            to Redis.
    """

    def __init__(
        self,
        host: str = "127.0.0.1",
        port: int = 6379,
        *,
        connection_pool=None,
        Reader: type = protocol.Resp3Reader,
        write_command: t.Callable = protocol.write_command,
    ):
        """Initialize the RedisClient.

        Arguments:
            host (str): The address to connect to (default: "127.0.0.1").
            port (int): The port to connect to (default: 6379).
            connection_pool: A connection pool to use. Overrides host and port.
                Leave as None to use the default ConnectionPool.
            Reader: the class to use for parsing replies from the server.
            write_command: the function used to prepare commands sent to the server.
        """
        self.host = host
        self.port = port

        if connection_pool is None:
            self.connection_pool = connections.ConnectionPool(host, port)
        else:
            self.connection_pool = connection_pool

        self.reader = Reader()
        self.write_command = write_command
        self.push_callbacks: defaultdict = defaultdict(list)

    async def receive(self, connection, push_only: bool = False):
        """Read the connection and return an object, calling any push callbacks.

        It is not recommended to call this directly. Use the :meth:`call` method,
        instead.

        Args:
            connection (trio.abc.Stream): The connection to read from.
            push_only (bool): If a push is received, return None
                and don't try to read anything else (default: False).

        Returns:
            The response from Redis, as parsed by the Reader class (or None, if
                push_only is True and a push is received).
        """
        while True:
            output = self.reader.get_object()
            if output is self.reader.sentinel:
                pass
            elif isinstance(output, protocol.RespPush):
                callbacks = self.push_callbacks[output.push_type]
                for callback in callbacks:
                    callback(output)
                if push_only:
                    return None
                continue  # pragma: nocover
            else:
                return output

            data = await connection.receive_some()
            self.reader.feed(data)

    async def send_command(self, command: bytes, *args: bytes, connection=None):
        """Send the given command to Redis and return the connection used.

        It is not recommended to call this directly. Use the :meth:`call` method,
        instead.

        Args:
            command (bytes): The command to send, such as b"PING" or b"SET".
            *args (bytes): The args to send with the command.
            connection (trio.abc.Stream): the connection to use, or None
                to get a connection from the pool.

        Returns:
            The connection used to send the command.

        """
        buffer = self.write_command(command, *args)
        if connection is None:
            connection = await self.connection_pool.wait_for_connection()
        await connection.send_all(buffer)
        return connection

    async def call(self, command: bytes, *args: bytes):
        """Send the given command to Redis and return the response.

        Args:
            command (bytes): The command to send, such as b"PING" or b"SET".
            *args (bytes): The args to send with the command.

        Returns:
            The response from Redis, as parsed by the Reader class.

        Example:
            call(b"SET", b"key_name", b"value") -> b"OK"
        """
        connection = await self.send_command(command, *args)
        response = await self.receive(
            connection, push_only=command.upper() in PUSH_COMMANDS
        )
        self.connection_pool.put_connection(connection)
        return response

    def register_push_callback(self, push_type: bytes, callback: t.Callable) -> None:
        """Register a function to be called when a push is received."""
        self.push_callbacks[push_type].append(callback)



"""The Midlevel client is a layer of abstraction above the lowlevel client.

In particular, this means:

1. Each command has its own function with type hints.

2. Where possible, arguments supplied to a command will be converted
   automatically from their type to the bytes that Redis expects to receive.

3. HELLO 3 is called automatically.

4. Some return types are changed, where it makes sense. For example, INCRBYFLOAT
   returns a float instead of the bytes that Redis returns.
"""
from .client import MidlevelClient



"""This module contains the MidlevelClient implementation.

All commands are divided by comments into sections based on https://redis.io/commands
"""
from redtrio.lowlevel import RedisClient


class MidlevelClient:
    """MidlevelClient is an abstraction on top of the lowlevel client.

    Args:
        **client_args: any arg accepted by :class:`lowlevel.RedisClient`.
    """

    def __init__(self, **client_args):
        """Initialize MidlevelClient."""
        self.client = RedisClient(**client_args)
        self.said_hello = False

    async def call(self, command: str, *args: str):
        """Send a command to Redis and return the response.

        Strings passed to this function will be converted to bytes.

        Args:
            command (str): The command to be sent, like "HELLO".
            *args (str): Arguments to be sent with the command.

        Returns:
            The response from Redis.
        """
        if not self.said_hello:
            self.said_hello = True
            await self.hello(3)
        encoded_args = [s.encode() for s in args]
        return await self.client.call(command.encode(), *encoded_args)

    async def hello(self, protocol: int) -> dict:
        """Say hello to Redis and let it know what protocol we're using.

        It's good to be friendly.

        Args:
            protocol (int): The protocol number. Currently 3 is the only option.

        Returns:
            A dict containing the response from Redis.
        """
        return await self.call("HELLO", str(protocol))

    ### Hash commands: https://redis.io/commands#hash ###
    async def hdel(self, key: str, *fields: str) -> int:
        """Implement the HDEL command (https://redis.io/commands/hdel)."""
        return await self.call("HDEL", key, *fields)

    async def hexists(self, key: str, field: str) -> int:
        """Implement the HEXISTS command (https://redis.io/commands/hexists)."""
        return await self.call("HEXISTS", key, field)

    async def hget(self, key: str, field: str) -> bytes:
        """Implement the HGET command (https://redis.io/commands/hget)."""
        return await self.call("HGET", key, field)

    async def hgetall(self, key: str) -> dict:
        """Implement the HGETALL command (https://redis.io/commands/hgetall)."""
        return await self.call("HGETALL", key)

    async def hincrby(self, key: str, field: str, increment: int) -> int:
        """Implement the HINCRBY command (https://redis.io/commands/hincrby)."""
        return await self.call("HINCRBY", key, field, str(increment))

    async def hincrbyfloat(self, key: str, field: str, increment: float) -> float:
        """Implement HINCRBYFLOAT command (https://redis.io/commands/hincrbyfloat).

        This command modifies the return response from `bytes` to `float`.

        Args:
            key (str): the key to increment.
            field (str): the field of the key to increment.
            increment (float): the amount to increment the field.

        Returns:
            The new value of the field (float).
        """
        return float(await self.call("HINCRBYFLOAT", key, field, str(increment)))

    async def hkeys(self, key: str) -> list:
        """Implement the HKEYS command (https://redis.io/commands/hkeys)."""
        return await self.call("HKEYS", key)

    async def hlen(self, key: str) -> int:
        """Implement the HLEN command (https://redis.io/commands/hlen)."""
        return await self.call("HLEN", key)

    async def hmget(self, key: str, *fields: str) -> list:
        """Implement the HMGET command (https://redis.io/commands/hmget)."""
        return await self.call("HMGET", key, *fields)

    async def hset(self, key: str, *args: str) -> str:
        """Implement the HSET command (https://redis.io/commands/hset)."""
        return await self.call("HSET", key, *args)

    async def hsetnx(self, key: str, field: str, value: str) -> int:
        """Implement the HSETNX command (https://redis.io/commands/hsetnx)."""
        return await self.call("HSETNX", key, field, value)

    async def hstrlen(self, key: str, field: str) -> int:
        """Implement the HSTRLEN command (https://redis.io/commands/hstrlen)."""
        return await self.call("HSTRLEN", key, field)

    async def hvals(self, key: str) -> list:
        """Implement the HVALS command (https://redis.io/commands/hvals)."""
        return await self.call("HVALS", key)

    ### String commands: https://redis.io/commands/#string ###
    async def append(self, key: str, value: str) -> bytes:
        """Implement the APPEND command (https://redis.io/commands/append)."""
        return await self.call("APPEND", key, value)

    async def get(self, key: str) -> bytes:
        """Implement the GET command (https://redis.io/commands/get)."""
        return await self.call("GET", key)

    async def set(
        self,
        key: str,
        value: str,
        *,
        ex: int = 0,
        px: int = 0,
        keepttl: bool = False,
        nx: bool = False,
        xx: bool = False,
    ):
        """Implement the SET command (https://redis.io/commands/set)."""
        command = ["SET", key, value]
        if bool(ex) + bool(px) + keepttl > 1:
            raise ValueError(
                f"More than one of {ex=}, {px=}, and {keepttl=} were specified"
            )
        if ex:
            command.extend(["EX", str(ex)])
        elif px:
            command.extend(["PX", str(px)])
        elif keepttl:
            command.append("KEEPTTL")

        if nx and xx:
            raise ValueError("Both nx and xx were specified")
        if nx:
            command.append("NX")
        elif xx:
            command.append("XX")

        return await self.call(*command)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['redtrio', 'redtrio.lowlevel', 'redtrio.midlevel']

package_data = \
{'': ['*']}

install_requires = \
['respy3>=0,<1', 'trio>=0.16,<1.0']

setup_kwargs = {
    'name': 'redtrio',
    'version': '0.5.0',
    'description': 'An async (Trio) client for Redis 6+',
    'long_description': '![Tests](https://github.com/Harrison88/redtrio/workflows/Tests/badge.svg)\n[![codecov](https://codecov.io/gh/Harrison88/redtrio/branch/master/graph/badge.svg)](https://codecov.io/gh/Harrison88/redtrio)\n[![PyPI](https://img.shields.io/pypi/v/redtrio.svg)](https://pypi.org/project/redtrio/)\n[![Documentation Status](https://readthedocs.org/projects/redtrio/badge/?version=latest)](https://redtrio.readthedocs.io/en/latest/?badge=latest)\n\nRedtrio is a modern async client for Redis, supporting the new RESP3 protocol.\n',
    'author': 'Harrison Morgan',
    'author_email': 'harrison.morgan@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Harrison88/redtrio',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Redtrio is a modern async client for Redis, supporting the new RESP3 protocol."""

from importlib import metadata

try:
    __version__ = metadata.version(__name__)
except Exception:  # pragma: no cover
    __version__ = "unknown"



"""The lowlevel package contains the bare minimum required for a Redis client.

Example:
    from redtrio.lowlevel import RedisClient
    client = RedisClient()
    result = await client.call(b"PING")
    print(result)
    b'PONG'

Modules:
    protocol - An implementation of Redis' RESP3 protocol
    connections - The default connection pool
    client - The Redis client, exported at the package level

Exports:
    RedisClient
"""

from .client import RedisClient
